// -*- Mode: c++ -*-
//
// (C) Copyright 1995-2005 OSR Open Systems Resources, Inc.
// All Rights Reserved.
//

//
// This header file describes the exposed data types for use with the
// file system wrapper interface.
//

#ifndef __OSR_WRAPPER_H__
#define __OSR_WRAPPER_H__ (1)

#ifndef _NTDDK_
#ifdef __cplusplus
extern "C" {
#endif
#include <ntddk.h>
#ifdef __cplusplus
};
#endif
#endif // _NTDDK_

//
// Per File System informtaion
//

typedef PVOID FS_VOL_HANDLE, *PFS_VOL_HANDLE;
typedef PVOID FS_FILE_HANDLE, *PFS_FILE_HANDLE;
typedef PVOID FS_LOCK_HANDLE, *PFS_LOCK_HANDLE;
typedef struct _FS_OPERATIONS *PFS_OPERATIONS;

//
// Directory entry formats have changed:
//
//  - Version 3 includes the file ID, for some of the directory types
//  - Version 2 includes the EA size
//  - Version 1 includes the base information
//
// Internally, the FSDK upgrades to the Version 3 format.  We encourage everyone
// to use the Version 3 format, but of course the earlier formats work as well.
//
//
typedef struct {
    ULONG EntrySize;

    //
    // We use a union here because this entry was initially mis-named.  The name should be
    // LastDirectoryEntry but to preserve compatibility we support the older name.
    // If this value is non-zero, the NEXT entry in the buffer is valid.
    //

    union {
        ULONG DirectoryEntryIndex;
        ULONG LastDirectoryEntry;
    };

    LARGE_INTEGER CreationTime;
    LARGE_INTEGER LastAccessTime;
    LARGE_INTEGER LastModifiedTime;
    LARGE_INTEGER Unused;
    LARGE_INTEGER ValidDataLength;
    LARGE_INTEGER DiskLength;  // storage-based
    ULONG Attributes;
    ULONG NameSize;
    UCHAR ShortNameSize;
    WCHAR ShortName[12];
    WCHAR Name[1]; // and for the balance of this entry
} FS_DIRECTORY_ENTRY, *PFS_DIRECTORY_ENTRY;

typedef struct {
    ULONG EntrySize;
    ULONG LastDirectoryEntry;
    LARGE_INTEGER CreationTime;
    LARGE_INTEGER LastAccessTime;
    LARGE_INTEGER LastModifiedTime;
    LARGE_INTEGER Unused;
    LARGE_INTEGER ValidDataLength;
    LARGE_INTEGER DiskLength;  // storage-based
    ULONG Attributes;
    ULONG NameSize;
    ULONG EaSize;
    UCHAR ShortNameSize;
    WCHAR ShortName[12];
    WCHAR Name[1]; // and for the balance of this entry
} FS_DIRECTORY_ENTRY2, *PFS_DIRECTORY_ENTRY2;

typedef struct {
    ULONG EntrySize;
    ULONG LastDirectoryEntry;
    LARGE_INTEGER CreationTime;
    LARGE_INTEGER LastAccessTime;
    LARGE_INTEGER LastModifiedTime;
    LARGE_INTEGER Unused;
    LARGE_INTEGER ValidDataLength;
    LARGE_INTEGER DiskLength;  // storage-based
    ULONG Attributes;
    ULONG NameSize;
    ULONG EaSize;
    UCHAR ShortNameSize;
    WCHAR ShortName[12];
    LARGE_INTEGER FileID;
    WCHAR Name[1]; // and for the balance of this entry
} FS_DIRECTORY_ENTRY3, *PFS_DIRECTORY_ENTRY3;

typedef struct {
    LARGE_INTEGER CreationTime;
    LARGE_INTEGER LastAccessTime;
    LARGE_INTEGER LastModifiedTime;
    ULONG Attributes;
    LARGE_INTEGER LastByteWritten; // a/k/a "ValidDataLength" on NT
    LARGE_INTEGER ValidDataLength; // a/k/a "EOF" on NT
    LARGE_INTEGER DiskLength;      // a/k/a "Allocation Size" on NT
    ULONG LinkCount;
    LARGE_INTEGER FileID;
} FS_FILE_ATTRIBUTES, *PFS_FILE_ATTRIBUTES;

typedef struct {
    LARGE_INTEGER CreationTime;
    ULONG SerialNumber;
    ULONG LabelLength;
    BOOLEAN Unused; // must be zero (or FALSE)
    LARGE_INTEGER VolumeSize; // in allocation units
    LARGE_INTEGER FreeSpace; // in allocation units
    ULONG BytesPerAllocationUnit;
    ULONG DeviceType; // FS type, one of DISK_FILE_SYSTEM, NETWORK_FILE_SYSTEM
    ULONG DeviceCharacteristics;
    WCHAR VolumeLabel[1]; // and for balance of entry
} FS_VOL_ATTRIBUTES, *PFS_VOL_ATTRIBUTES;

typedef struct {
    LARGE_INTEGER CreationTime;
    ULONG SerialNumber;
    ULONG LabelLength;
    BOOLEAN Flags;
    BOOLEAN Unused[3];
    LARGE_INTEGER VolumeSize; // in allocation units
    LARGE_INTEGER FreeSpace; // in allocation units
    ULONG BytesPerAllocationUnit;
    ULONG DeviceType; // FS type, one of DISK_FILE_SYSTEM, NETWORK_FILE_SYSTEM
    ULONG DeviceCharacteristics;
    USHORT MaximumComponentNameLength;
    USHORT FileSystemNameLength; 
    WCHAR VolumeLabel[1]; // and for balance of entry
    // file system name immediately follows the volume label
} FS_EXTENDED_VOL_ATTRIBUTES, *PFS_EXTENDED_VOL_ATTRIBUTES;

#define OW_VOL_ATTR_FLAG_EXTENDED (0x01)

typedef struct {
    ULONG EntrySize;
    ULONG NameSize;
    LARGE_INTEGER ValidDataLength;
    LARGE_INTEGER DiskLength;
    WCHAR Name[1];
} FS_STREAM_ENTRY, *PFS_STREAM_ENTRY;

typedef UCHAR FS_TYPE;

typedef struct {
    LARGE_INTEGER VolumeOffset;
    PMDL Buffer;
    ULONG BufferSize;
} OW_PSEUDO_VOLUME_IO_INFO, *POW_PSEUDO_VOLUME_IO_INFO;

typedef struct {
    ULONG Bytes;
} OW_PSEUDO_VOLUME_IO_COMPLETE, *POW_PSEUDO_VOLUME_IO_COMPLETE;

#define OW_FS_TYPE_NULL    (0x00)
#define OW_FS_TYPE_MEDIA   (0x10)
#define OW_FS_TYPE_NETWORK (0x20)
#define OW_FS_TYPE_PSEUDO  (0x30)

#define OW_FS_TYPE(x) (x & 0xF0)

//
// Modifiers for FS types.  Note that they may be ORed
// together.
//

#define OW_FS_TYPE_MEDIA_DISK  (0x01)
#define OW_FS_TYPE_MEDIA_TAPE  (0x02)
#define OW_FS_TYPE_MEDIA_CDROM (0x04)

#define OW_FS_TYPE_NETWORK_NAMED_PIPE (0x01)
#define OW_FS_TYPE_NETWORK_MAILSLOT   (0x02)

//
// If used, this will cause the device object
// to be declared as type FILE_DEVICE_DISK_FILE_SYSTEM
// rather than FILE_DEVICE_FILE_SYSTEM.
//
#define OW_FS_TYPE_PSEUDO_DISK (0x01)
#define OW_FS_TYPE_PSEUDO_NETWORK (0x08)

#define OW_HANDLE_TYPE_FILE          0x01
#define OW_HANDLE_TYPE_DIRECTORY     0x02
#define OW_HANDLE_TYPE_VOLUME        0x03
#define OW_HANDLE_TYPE_SYMBOLIC_LINK 0x04
#define OW_HANDLE_TYPE_PRINTER       0x05
#define OW_HANDLE_TYPE_REPARSE_POINT 0x06
#define OW_HANDLE_TYPE_OTHER         0xFF

//
// Backwards compatibility
//

#define OSRFS_TYPE_FILE OW_HANDLE_TYPE_FILE
#define OSRFS_TYPE_DIRECTORY OW_HANDLE_TYPE_DIRECTORY
#define OSRFS_TYPE_SYMBOLIC_LINK OW_HANDLE_TYPE_SYMBOLIC_LINK
#define OSRFS_TYPE_OTHER OW_HANDLE_TYPE_OTHER

//
// file attributes match 1-1 with the standard NT file attributes
//
// FILE_ATTRIBUTE_*
//

//
// volume attributes match 1-1 with the standard NT volume attributes
//
#define FILE_CASE_SENSITIVE_SEARCH      0x00000001  
#define FILE_CASE_PRESERVED_NAMES       0x00000002  
#define FILE_UNICODE_ON_DISK            0x00000004  
#define FILE_PERSISTENT_ACLS            0x00000008  
#define FILE_FILE_COMPRESSION           0x00000010  
#define FILE_VOLUME_IS_COMPRESSED       0x00008000  

//
// Device characteristics map 1-1 with their DDK counterparts:
//
// FILE_REMOVABLE_MEDIA
// FILE_READ_ONLY_DEVICE
// FILE_FLOPPY_DISKETTE
// FILE_WRITE_ONCE_MEDIA
// FILE_REMOTE_DEVICE
// FILE_DEVICE_IS_MOUNTED
// FILE_VIRTUAL_VOLUME
//


//
// The FSDK can be used either as a DLL or as a library.  If it is used
// as a library, OSR_WRAPPER_API is a null declaration.  If it is used as a
// DLL, OSR_WRAPPER_API is declared as a dllimport type.
//

#if !defined(OSR_WRAPPER_API)

#if defined(FSDK_LIBRARY)

#define OSR_WRAPPER_API

#else

#define OSR_WRAPPER_API _declspec(dllimport)

#endif // FSDK_LIBRARY

#endif // OSR_WRAPPER_API

//
// The following value is the current version
//
//  Use the high byte to indicate a custom version number
//
//

#define OSR_WRAPPER_VERSION_1_0            (0x00000004)
#define OSR_WRAPPER_VERSION_1_1            (0x00000005)
#define OSR_WRAPPER_VERSION_1_2            (0x00000006)
#define OSR_WRAPPER_VERSION_2_0_PRERELEASE (0x0000000F)
#define OSR_WRAPPER_VERSION_2_0            (0x00000010)

#define OSR_WRAPPER_VERSION OSR_WRAPPER_VERSION_2_0_PRERELEASE

//
// FSDK 2.0 is not binary compatible with FSDK 1.x.
//
#define OSR_VALID_VERSION(x) ((OSR_WRAPPER_VERSION_2_0_PRERELEASE == x) || (OSR_WRAPPER_VERSION_2_0 == x))

//
// Exported API functions.
//

#ifdef __cplusplus
extern "C" {
#endif

//
// OwRegister
//
//   This routine is used to register the new file system with the FSDK
//
// Inputs:
//   FsdDriverObject - this is the driver object for the file system
//   RegistryPath - this is the path to the FSD in the registry
//   VersionNumber - normally, OSR_WRAPPER_VERSION (and hence a compile-time constant)
//                   this value is used for version checking
//   FsdType - one of the file system types
//   FsdOperations - the operations and configuration for this file system
//
// Outputs:
//   RegistrationHandle - the handle to use for a subsequent deregistration
//
// Returns:
//   STATUS_REVISION_MISMATCH - the VersionNumber specified is incompatible with the
//                              loaded version of the FSDK.
//   STATUS_INVALID_PARAMETER_4 - the type of file system specified was invalid
//   STATUS_INSUFFICIENT_RESOURCES - a memory allocation operation failed
//   STATUS_ACCESS_VIOLATION - a caller-provided buffer was invalid
//   STATUS_DEVICE_ALREADY_ATTACHED - the NULL file system has already been registered
//   STATUS_SUCCESS - the file system loaded successfully
//
// Notes:
//   This function initializes the driver object for the file system.  Thus, handling of
//   I/O requests is done by calling FSDK functions, rather than calling functions within
//   the FSD directly.  The FSDK will then call FSD functions (via the FsdOperations
//   vector) as necessary.
//

    NTSTATUS
    OSR_WRAPPER_API
    OwRegister(
        IN PDRIVER_OBJECT FsdDriverObject,
        IN PUNICODE_STRING RegistryPath,
        IN ULONG VersionNumber,
        IN FS_TYPE FsdType,
        IN PFS_OPERATIONS FsdOperations,
        OUT PVOID *RegistrationHandle);

//
// OwDeregister
//
//   This routine is used to deregister a previously registered file system with the FSDK.
//
// Inputs:
//   RegistrationHandle - this is the registration handle provided by OwRegister
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_INVALID_DEVICE_REQUEST - the file system has active volumes and cannot be deregistered.
//   STATUS_SUCCESS - the file system is deregistered.
//
// Notes:
//   None.
//

    NTSTATUS
    OSR_WRAPPER_API
    OwDeregister(
        IN PVOID RegistrationHandle);

//
// OwFlushCache
//
//   This routine is used to ensure any modified data in the VM cache is written to the file system.
//
// Inputs:
//   FileHandle - the file being flushed
//   Purge - if TRUE, this call will purge the VM system (if possible)
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_FILE_LOCK_CONFLICT - FSDK internal locks cannot be acquired, the file was not flushed
//   STATUS_USER_MAPPED_FILE - the file is mapped by an application and cannot be purged (it was flushed)
//   STATUS_SUCCESS - the file was flushed (and purged, if requested)
//   I/O errors - any I/O errors that occur may be returned at this point.
//
// Notes:
//   It is not possible for the FSDK to purge data in all cases, notably those where the file is in use
//   by an application.  In such a case, the FSDK will flush the data but indicates an unsuccessful purge
//   condition.
//
//   Due to the complexities of the synchronization model (this is an upcall into the file system) it may not
//   be possible for the FSDK to obtain the lock necessary to purge the volume.  In such a case, the FSDK will
//   indicate the failure to the FSD.  The FSD should post a work item and retry the operation in a different
//   thread context (where it is safe to block.)
// 

    NTSTATUS
    OSR_WRAPPER_API
    OwFlushCache(
        IN FS_FILE_HANDLE FileHandle,
        IN BOOLEAN Purge);
    
//
// OwPurgeCache
//
//   This routine is used to ask the FSDK to discard any cached data.  For files this includes data
//   cached in the VM system.  For directories, this includes any cached directory state.
//
// Inputs:
//   FileHandle - the file that we are going to purge.
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_INVALID_HANDLE - the FileHandle provided does not reference a valid file
//   STATUS_UNSUCCESFUL - the file could not be purged (VM system failed the request)
//   STATUS_SUCCESS - the purge operation was successful
//
// Notes:
//   Purge sounds like a simple function but in fact turns out to be quite complicated because
//   the act of purging can (and frequently does) cause re-entrancy from the VM system back
//   into the file system.  Because of this a purge may fail because the VM system cannot obtain
//   the necessary internal synchronization locks it needs to perform the operation.
//
//   If this fails, we suggest retrying the operation via a work routine.
//

    NTSTATUS
    OSR_WRAPPER_API
    OwPurgeCache(
        IN FS_FILE_HANDLE FileHandle);

//
// OwDisableFileCache
//
//  This routine is used to disable caching for the given file.  For directories, this
//  means that the directory enumeration will not be cached (assuming DisableReadCaching is TRUE).
//
// Inputs:
//   FileHandle - the file being modified
//   DisableReadCaching - indicates if this is write-through (no write caching) or no caching
//                        (no read caching OR write caching.)
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_INVALID_HANDLE - the FileHandle was not valid
//   STATUS_USER_MAPPED_FILE - the file in question is in use by an application
//   STATUS_SUCCESS - caching has been adjusted for this file
//
// Notes:
//   This routine is used by an FSD to disable caching of data for this file.  Files that are 
//   memory mapped and compressed files cannot be accessed without caching.
//
//   An FSD that disables caching for a file must be prepared to handle I/O operations from the
//   user applications, including unaligned buffers, non-integral lengths, etc.  The FSDK does
//   not isolate the FSD from these conditions for non-cached I/O.
//
//   This function has no effect for anything other than a file or directory (symbolic link,
//   reparse point, etc.)
//

    NTSTATUS
    OSR_WRAPPER_API
    OwDisableFileCache(
        IN FS_FILE_HANDLE FileHandle,
        IN BOOLEAN DisableReadCaching);

//
// OwEnableFileCache
//
//   This routine is used to enable caching for the given file.  For directories, this
//   means that the directory enumeration will be cached.
//
// Inputs:
//   FileHandle - the file being modified
//   EnableWriteCaching - the FSDK may cache dirty data (versus write-through caching)
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_INVALID_HANDLE - the file handle provided was not valid
//   STATUS_SUCCESS - caching has been enabled for the file.
//
// Notes:
//   None.
//

    NTSTATUS
    OSR_WRAPPER_API
    OwEnableFileCache(
        IN PVOID FileHandle,
        IN BOOLEAN EnableWriteCaching);

//
// OwDisableVolumeCache
//
//   This routine is used to disable caching for all newly opened files and directories for
//   the given volume.
//
// Inputs:
//   FsdVolumeHandle - the handle of the volume in question
//   DisableReadCaching - indicates if this is write-through (no write caching) or no caching
//                        (no read caching OR write caching.)
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_INVALID_HANDLE - the volume handle provided was not valid
//   STATUS_SUCCESS - Caching was disabled
//
// Notes:
//   A network file system may disable caching on the UNC pseudo volume (that is, all
//   files opened without a drive letter connection) by passing in the UNC root directory
//   handle rather than a volume handle.
//
//   This routine is used by an FSD to disable caching of data for this file.  Files that are 
//   memory mapped and compressed files cannot be accessed without caching.
//
//   An FSD that disables caching for a file must be prepared to handle I/O operations from the
//   user applications, including unaligned buffers, non-integral lengths, etc.  The FSDK does
//   not isolate the FSD from these conditions for non-cached I/O.
//
//   This function has no effect for anything other than a file or directory (symbolic link,
//   reparse point, etc.)
//
//   Note that the change dictated by this call will apply to newly opened files and directories
//   and will not modify the behavior of files or directories previously opened on the volume.
//  

    NTSTATUS
    OSR_WRAPPER_API
    OwDisableVolumeCache(
        IN  FS_VOL_HANDLE FsdVolumeHandle,
        IN BOOLEAN DisableReadCaching);

//
// OwEnableVolumeCache
//
//   This routine is used to enable caching for newly opened files/directories on the given volume.
//
// Inputs:
//   FsdVolumeHandle - the handle of the volume in question
//   EnableWriteCaching - the FSDK may cache dirty data (versus write-through caching)
//
// Outputs:
//   None.
//
// Returns:
//   STATUS_INVALID_HANDLE - the volume handle provided was not valid
//   STATUS_SUCCESS - Caching was disabled
//
// Notes:
//   None.
//


    NTSTATUS
    OSR_WRAPPER_API
    OwEnableVolumeCache(
        IN  FS_VOL_HANDLE FsdVolumeHandle,
        IN BOOLEAN EnableWriteCaching);

//
// OwSetReadAhead
//
//   This routine sets the read ahead size for the given file.
//
// Inputs:
//   FileHandle - this is the file handle being controlled
//   NumberOfPages - this is the number of pages to perform as a read-ahead size
//
// Outputs:
//   None.
//
// Returns:
//   Void function
//
// Notes:
//   Setting the read ahead size to an arbitrarily large number does not mean that
//   the VM system will use that size - typically it will divide up a large read-ahead
//   and perform it as a series of smaller I/O operations.
//
//   This has no effect if the FileHandle refers to anything other than a file.

    VOID
    OSR_WRAPPER_API
    OwSetReadAhead(
        IN FS_FILE_HANDLE FileHandle,
        IN UCHAR NumberOfPages);

//
// OwSetWriteBehind
//
//   This routine sets the write behind cache size for the given file.
//
// Inputs:
//   FileHandle - this is the file handle for the file in question
//   NumberOfPages - the maximum number of (dirty) pages to allow in memory
//
// Outputs:
//   None.
//
// Returns:
//   Void function
//
// Notes:
//   The default behavior is to allow an arbitrary number of dirty pages to
//   be cached; the lazy write will flush data based upon its age.  By setting
//   a maximum threshold, any attempts to write into the cache that cause this
//   threshold to be exceeded force the application to block and I/O to be
//   performed before the application can proceed.


    VOID
    OSR_WRAPPER_API
    OwSetWriteBehind(
        IN FS_FILE_HANDLE FileHandle,
        IN UCHAR NumberOfPages);


//
// OwGetMediaDeviceObject
//
//   This routine is used to convert the media handle to an NT device object.
//
// Inputs:
//   MediaHandle - the media on which this operation is being performed.
//
// Outputs:
//   DeviceObject - the NT device object corresponding to the MediaHandle.
//
// Returns:
//   STATUS_INVALID_HANDLE - the MediaHandle provided was not a valid FSDK media handle.
//   STATUS_SUCCESS - the conversion was successful
//
// Notes:
//   This routine is normally used by an FSD that does not wish to use the FSDK
//   I/O routines.  The FSDK does not require that file systems use its media I/O
//   interface.
//

    NTSTATUS
    OSR_WRAPPER_API
    OwGetMediaDeviceObject(IN PVOID MediaHandle,
                           OUT PDEVICE_OBJECT *DeviceObject);


//
// OwGetPseudoDeviceObject
//
//  This routine is used to convert the pseudo volume handle to the corresponding
//  NT device object (used to represent that pseudo volume.)
//
// Inputs:
//  PseudoVolumeHandle - this is the pseudo volume handle, provided to the FSD by
//                       the FSDK.
//
// Outputs:
//  DeviceObject - this is the device representing the pseudo volume.
//
// Returns:
//  STATUS_INVALID_HANDLE - the PseudoVolumeHandle provided was not a valid FSDK pseudo
//                          volume handle.
//  STATUS_SUCCESS - the conversion was successful.
//
// Notes:
//  This routine is normally used by an FSD that needs to be able to call the FSDK on
//  behalf of the specific pseudo volume.

    NTSTATUS
    OSR_WRAPPER_API
    OwGetPseudoDeviceObject(IN PVOID PseudoVolumeHandle,
                            OUT PDEVICE_OBJECT *DeviceObject);


//
// OwGetTopLevelIrp
//
//  This entry point can be used by an FSD to retrieve the top level IRP currently in
//  progress.  For read/write this is meaningless, but for most others it is fine.
// 
//
// Inputs:
//  None.
//
// Outputs:
//  None.
//
// Returns:
//  Pointer to the top level irp (or null if there isn't one)
//  
// Notes:
//  None.
//

    PIRP
    OSR_WRAPPER_API
    OwGetTopLevelIrp(void);


//
// OwNotifyDirectoryChange
//
//  This is a mechanism whereby the FSD can notify us of an external
//  event which triggers an FSD change.
//
// Inputs:
//  FileHandle - the file handle upon which we are operating
//  FullTargetName - the path name of the target (the file that changed
//  TargetNameOffset - the offset in the FullTargetName to the "final" component of the name
//  StreamName - optional, indicates if this was a STREAM of a file which changed
//  ParentName - optional, indicates the name of the DIRECTORY that changed (for multi-linked directories)
//  FilterMatch - indicates WHAT changed
//
// Outputs:
//  None.
//
// Returns:
//  STATUS_SUCCESS - it worked
//  STATUS_INVALID_DEVICE_STATE - IRQL > PASSIVE_LEVEL
//  STATUS_INVALID_HANDLE - the handle provided was not valid
//  other errors as appropriate
//
// Notes:
//  It is OK to notify on a directory nobody cares about - that returns SUCCESS.  The
//  handle must be valid, though, etc.
//


    NTSTATUS
    OSR_WRAPPER_API
    OwNotifyDirectoryChange(IN FS_FILE_HANDLE FileHandle,
                            IN PUNICODE_STRING FullTargetName,
                            IN USHORT TargetNameOffset,
                            IN PUNICODE_STRING StreamName,
                            IN PUNICODE_STRING ParentName,
                            IN ULONG FilterMatch);

//
// OwPostWork
//
//  This may be used by an FSD to request that an FSD-provided function be called
//  in a worker thread to perform a background operation.  This will be accomplished
//  by using one of the FSDK internal work queues.
//
// Inputs:
//   WorkToDo - the work to perform (by calling the specified function within this parameter)
//
// Outputs:
//   None.
//
// Returns:
//   Void function
//
// Notes:
//   This interface provides an alternative to the system queues.
//

    typedef struct _OW_WORK {
        VOID        (*WorkFunction)(PVOID ContextValue);
        PVOID       ContextValue;
    } OW_WORK, *POW_WORK;
    
    VOID
    OSR_WRAPPER_API
    OwPostWork(POW_WORK WorkToDo);

//
// OwAllocateSDBuffer
//
//   This routine is used to allocate a security descriptor buffer
//
// Inputs:
//   BufferSize - the size of the buffer needed for the Security Descriptor (SD)
//
// Outputs:
//   Buffer - this is the allocated buffer (can be NULL)
//
// Returns:
//   STATUS_INSUFFICIENT_RESOURCES - the buffer could not be allocated
//   STATUS_SUCCESS - the buffer has been allocated
//   STATUS_INVALID_BUFFER_SIZE - the buffer size was zero
//
// Notes:
//   There is no equivalent "free" because the FSDK will handle freeing
//   the security descriptor.
//  

    NTSTATUS
    OSR_WRAPPER_API
    OwAllocateSDBuffer(IN ULONG BufferSize, 
                       OUT PVOID *Buffer);


#ifdef __cplusplus
};
#endif

//
// FSD Operation declarations
//

typedef
NTSTATUS
(*FS_MOUNT)(
    IN PVOID MediaHandle,
    IN OUT PFS_VOL_HANDLE FsdVolumeHandle,
    OUT PFS_FILE_HANDLE RootDirectoryHandle);

typedef
NTSTATUS
(*FS_VERIFY)(
    IN FS_VOL_HANDLE FsdVolumeHandle);

typedef
NTSTATUS
(*FS_UNMOUNT)(
    IN FS_VOL_HANDLE FsdVolumeHandle);

typedef
NTSTATUS
(*FS_FSCTRL)(
    IN FS_VOL_HANDLE FsdVolumeHandle,
    IN ULONG ControlCode,
    IN PVOID InputBuffer,
    IN ULONG InputBufferSize,
    IN PVOID OutputBuffer,
    IN ULONG OutputBufferSize,
    OUT PULONG BytesWritten,
    OUT PULONG BytesNeeded);

typedef
NTSTATUS
(*FS_GET_QUOTA)(
    IN FS_VOL_HANDLE FsdVolumeHandle,
    IN OUT PVOID QuotaInformationBuffer,
    IN OUT PULONG QuotaInformationBufferSize);

typedef
NTSTATUS
(*FS_SET_QUOTA)(
    IN FS_VOL_HANDLE FsdVolumeHandle,
    IN PVOID QuotaInformationBuffer,
    IN ULONG QuotaInformationBufferSize);

typedef
NTSTATUS
(*FS_GET_VOLUME_ATTRIBUTES)(
    IN FS_VOL_HANDLE FsdVolumeHandle,
    IN OUT PFS_VOL_ATTRIBUTES VolumeAtttributesBuffer,
    IN OUT PULONG VolumeAttributesBufferSize);

typedef
NTSTATUS
(*FS_SET_VOLUME_ATTRIBUTES)(
    IN FS_VOL_HANDLE FsdVolumeHandle,
    IN PFS_VOL_ATTRIBUTES VolumeAtttributes);

typedef
NTSTATUS
(*FS_LOOKUP)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN PWSTR FileName,
    IN BOOLEAN CaseInsensitive,
    OUT PFS_FILE_HANDLE FileHandle,
    OUT PULONG Type);

typedef
NTSTATUS
(*FS_LOOKUP_BY_ID)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN LARGE_INTEGER FileId,
    OUT PFS_FILE_HANDLE FileHandle,
    OUT PULONG Type);

typedef
NTSTATUS
(*FS_LOOKUP_BY_OBJECT_ID)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN GUID *Guid,
    OUT PFS_FILE_HANDLE FileHandle,
    OUT PULONG Type);

typedef
NTSTATUS
(*FS_LOOKUP_PATH)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN PWSTR PathName,
    IN BOOLEAN CaseInsensitive,
    IN PIO_SECURITY_CONTEXT CallerSecurityContext,
    OUT PFS_FILE_HANDLE ParentDirectoryHandle,
    OUT PFS_FILE_HANDLE FileHandle,
    OUT PULONG Type);


typedef
NTSTATUS
(*FS_SET_ACCESS_MODE)(
    IN FS_FILE_HANDLE FileHandle,
    IN UCHAR AccessMode);

typedef
NTSTATUS
(*FS_RELEASE)(
    IN FS_FILE_HANDLE FileHandle);

typedef
NTSTATUS
(*FS_CREATE)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN PWSTR FileName,
    IN PWSTR ShortFileName,
    IN ULONG FileType,
    OUT PFS_FILE_HANDLE NewFileHandle);

typedef
NTSTATUS
(*FS_DELETE)(
    IN FS_FILE_HANDLE FileHandle);

typedef
NTSTATUS
(*FS_DELETE2)(
    IN FS_FILE_HANDLE FileHandle,
    IN FS_FILE_HANDLE ParentDirectoryHandle,
    IN PWSTR FileName);

typedef
NTSTATUS
(*FS_RENAME)(
    IN FS_FILE_HANDLE SourceDirectoryHandle,
    IN FS_FILE_HANDLE TargetDirectoryHandle,
    IN FS_FILE_HANDLE SourceFileHandle,
    IN PWSTR TargetFileName,
    IN PWSTR TargetFileAlternateName,
    OUT PFS_FILE_HANDLE TargetFileHandle);

typedef
NTSTATUS
(*FS_RENAME2)(
    IN FS_FILE_HANDLE SourceDirectoryHandle,
    IN FS_FILE_HANDLE TargetDirectoryHandle,
    IN FS_FILE_HANDLE SourceFileHandle,
    IN PWSTR SourceFileName,
    IN PWSTR TargetFileName,
    IN PWSTR TargetFileAlternateName,
    OUT PFS_FILE_HANDLE TargetFileHandle);

typedef
NTSTATUS
(*FS_ACCESS)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PACCESS_STATE AccessState);

typedef
NTSTATUS
(*FS_GET_SECURITY)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PVOID* SecurityDescriptor,
    IN OUT PULONG SecurityDescriptorSize);

typedef
NTSTATUS
(*FS_SET_SECURITY)(
    IN FS_FILE_HANDLE FileHandle,
    IN PVOID SecurityDescriptor,
    IN ULONG SecurityDescriptorSize);

typedef
NTSTATUS
(*FS_LINK)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN FS_FILE_HANDLE FileHandle,
    IN PWSTR LinkName,
    IN PWSTR ShortLinkName);

typedef
NTSTATUS
(*FS_READ_SYMBOLIC_LINK)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PWSTR SymbolicLinkNameBuffer,
    IN OUT PULONG SymbolicLinkNameBufferSize);

typedef
NTSTATUS
(*FS_WRITE_SYMBOLIC_LINK)(
    IN FS_FILE_HANDLE FileHandle,
    IN PWSTR SymbolicLinkNameBuffer,
    IN ULONG SymbolicLinkNameBufferSize);

typedef
NTSTATUS
(*FS_GET_ATTRIBUTES)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PFS_FILE_ATTRIBUTES FileAtttributes);

typedef
NTSTATUS
(*FS_SET_ATTRIBUTES)(
    IN FS_FILE_HANDLE FileHandle,
    IN PFS_FILE_ATTRIBUTES FileAtttributes);

typedef
NTSTATUS
(*FS_FLUSH)(
    IN FS_FILE_HANDLE FileHandle);

typedef
NTSTATUS
(*FS_SET_LENGTH)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER AllocationSize,
    IN LARGE_INTEGER DataSize);

typedef
NTSTATUS
(*FS_GET_EA)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PVOID EABuffer,
    IN OUT PULONG EABufferSize);

typedef
NTSTATUS
(*FS_SET_EA)(
    IN FS_FILE_HANDLE FileHandle,
    IN PVOID EABuffer,
    IN ULONG EABufferSize);

typedef
NTSTATUS
(*FS_READ)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER FileOffset,
    IN PMDL Buffer,
    IN ULONG BufferSize,
    OUT PULONG BytesRead);

typedef
NTSTATUS
(*FS_WRITE)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER FileOffset,
    IN PMDL Buffer,
    IN ULONG BufferSize,
    OUT PULONG BytesWritten);

typedef
NTSTATUS
(*FS_IOCTL)(
    IN FS_FILE_HANDLE FileHandle,
    IN ULONG ControlCode,
    IN PVOID InputBuffer,
    IN ULONG InputBufferSize,
    IN PVOID OutputBuffer,
    IN ULONG OutputBufferSize,
    OUT PULONG BytesWritten,
    OUT PULONG BytesNeeded);

typedef
NTSTATUS
(*FS_READ_DIRECTORY)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN OUT PULONG DirectoryOffset,
    IN PFS_DIRECTORY_ENTRY OutputBuffer,
    IN ULONG OutputBufferSize);

typedef
NTSTATUS
(*FS_READ_DIRECTORY2)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN OUT PULONG DirectoryOffset,
    IN PFS_DIRECTORY_ENTRY2 OutputBuffer,
    IN ULONG OutputBufferSize);

typedef
NTSTATUS
(*FS_READ_DIRECTORY3)(
    IN FS_FILE_HANDLE DirectoryHandle,
    IN OUT PULONG DirectoryOffset,
    IN PFS_DIRECTORY_ENTRY3 OutputBuffer,
    IN ULONG OutputBufferSize);

typedef
NTSTATUS
(*FS_READ_STREAM_INFORMATION)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PFS_STREAM_ENTRY OutputBuffer,
    IN OUT PULONG OutputBufferSize);

typedef
NTSTATUS
(*FS_QUERY_PATH)(
    IN PWSTR PathName,
    OUT PULONG PathNameAccepted,
    OUT PFS_FILE_HANDLE DirectoryHandle);

typedef
NTSTATUS
(*FS_CONNECT)(
    IN PWSTR RemoteName,
    IN UCHAR ConnectionType,
    OUT PFS_VOL_HANDLE FsdVolumeHandle,
    OUT PFS_FILE_HANDLE RootDirectoryHandle);

typedef
NTSTATUS
(*FS_DISCONNECT)(
    IN FS_VOL_HANDLE FsdVolumeHandle);

typedef
VOID
(*FS_UNC_ROOT)(
    OUT PFS_FILE_HANDLE DirectoryHandle);

typedef
NTSTATUS
(*FS_LOCK)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER Offset,
    IN LARGE_INTEGER Length,
    IN BOOLEAN Exclusive,
    IN BOOLEAN Wait,
    OUT PFS_LOCK_HANDLE LockHandle);

typedef
NTSTATUS
(*FS_UNLOCK)(
    IN FS_FILE_HANDLE FileHandle,
    IN FS_LOCK_HANDLE LockHandle);

typedef
NTSTATUS
(*FS_GET_UNC_VOLUME_ATTRIBUTES)(
    IN FS_FILE_HANDLE,
    IN PFS_VOL_ATTRIBUTES VolumeAtttributes,
    IN OUT PULONG VolumeAttributesBufferSize);


typedef
NTSTATUS
(*FS_CHECK_LOCK)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER FileOffset,
    IN ULONG BufferSize,
    IN BOOLEAN Read);

typedef
NTSTATUS
(*FS_UPDATE_SHARE_ACCESS)(
    IN FS_FILE_HANDLE FileHandle,
    IN ULONG ShareAccess);

typedef
NTSTATUS
(*FS_REMOVE_SHARE_ACCESS)(
    IN FS_FILE_HANDLE FileHandle,
    IN ULONG ShareAccess);

//
// FS_OPEN - warning, this interface is experimental and subject
//           to change.  Comments are welcome, please send them
//           to bugs@osr.com.  
//           
typedef
NTSTATUS
(*FS_OPEN)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PACCESS_STATE AccessState,
    IN ULONG ShareAccess,
    OUT PBOOLEAN ReadCache,
    OUT PBOOLEAN WriteBack
    );

//
// FS_CLOSE - warning, this interface is experimental and subject
//            to change.  Comments are welcome, please send them
//            to bugs@osr.com.
//
typedef
NTSTATUS
(*FS_CLOSE)(
    IN FS_FILE_HANDLE FileHandle,
    IN ACCESS_MASK GrantedAccess,
    IN ULONG ShareAccess,
    OUT PBOOLEAN FlushCache,
    OUT PBOOLEAN PurgeCache
);

//
// FS_READ_COMPRESSED
//
//  warning, this interface is experimental and subject
//  to change.  Comments are welcome, please send them
//  to bugs@osr.com.
//

typedef
NTSTATUS
(*FS_READ_COMPRESSED)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER FileOffset,
    IN OUT PMDL Buffer,
    IN ULONG BufferSize,
    IN OUT PVOID CompressionInfo,
    IN ULONG CompressionInfoLength,
    OUT PULONG BytesRead
);


//
// FS_WRITE_COMPRESSED
//
//  warning, this interface is experimental and subject
//  to change.  Comments are welcome, please send them
//  to bugs@osr.com.

typedef
NTSTATUS
(*FS_WRITE_COMPRESSED)(
    IN FS_FILE_HANDLE FileHandle,
    IN LARGE_INTEGER FileOffset,
    IN PMDL Buffer,
    IN ULONG BufferSize,
    IN PVOID CompressionInfo,
    IN ULONG CompressionInfoLength,
    OUT PULONG BytesWritten
);

//
// FS_CLEAR
//
//  warning, this interface is experimental and subject
//  to change.  Comments are welcome, please send them
//  to bugs@osr.com.
//  
typedef
NTSTATUS
(*FS_CLEAR)(
    IN FS_FILE_HANDLE FileHandle,
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length
);

//
// FS_GET_NAMES
//
// Inputs:
//  FileHandle - the file for which we are retrieving names
//
// Outputs:
//  LongName - the long name for this file
//  ShortName -the short name for this file
//
// Returns:
//  STATUS_SUCCESS - the call succeeded
//  STATUS_BUFFER_OVERFLOW - one (or both) of the buffers was too small, Length is set appropriately
//
// Notes:
//  The FSDK will set the MaximumLength field to indicate the size of the buffer provided within
//  the UNICODE_STRING structures.
//
//  If there is only a short name, return Length = 0 for the long name.
//  Only those file systems that support short and long names need to support this; otherwise the
//  FSDK will assume the file (or link) has only a single canonical name.
//
typedef
NTSTATUS
(*FS_GET_NAMES)(
    IN FS_FILE_HANDLE FileHandle,
    IN OUT PUNICODE_STRING LongName,
    IN OUT PUNICODE_STRING ShortName
);

//
// FS_SHUTDOWN
//
//  Warning: this interface is experimental and subject
//  to change.  Comments are welcome, please send them to
//  bugs@osr.com
//
// Inputs:
//  Stage - OW_SHUTDOWN_INITIATE (start of shutdown), OW_SHUTDOWN_COMPLETE (termination of shutdown)
//
// Outputs:
//  None.
//
// Returns:
//  ignored
//
// Notes:
//  The FSDK is advising the underlying driver that a shutdown is in progress.  It calls
//  initially to notify that shutdown processing has initiated; the FSDK then dismounts
//  all volumes, and then finally it notifies the FSD that shutdown processing has completed.
//
typedef
NTSTATUS
(*FS_SHUTDOWN)(ULONG Stage);

#define OW_SHUTDOWN_INITIATE 0
#define OW_SHUTDOWN_COMPLETE 1


//
// FS_RESERVED_FUNCTION
//
//  This entry type is reserved for future expansion and must
//  be zero.
//
//
typedef
NTSTATUS
(*FS_RESERVED_FUNCTION)(void);


typedef struct _FS_OPERATIONS_FILTER {

    //
    // Indicates the verison of this structure
    //
    USHORT           Version;

    //
    // Indicates the size of this structure in bytes.
    // This allows the filter table to grow without
    // sacrificing backwards compatibility.
    //
    USHORT           Size;


    //
    // These entry points are called in the FSD before
    // the FSDK gains control.   This allows the FSD
    // to override FSDK behavior if that's desireable.
    //
    PDRIVER_DISPATCH DispatchFilters[IRP_MJ_MAXIMUM_FUNCTION+1];

    //
    // These entry points are called in the FSD after the FSDK
    // is done processing the I/O request.  This allows the FSD
    // an opportunity to change the completion behavior of this
    // I/O request if desired.
    //
    PDRIVER_DISPATCH CompletionFilters[IRP_MJ_MAXIMUM_FUNCTION+1];

} FS_OPERATIONS_FILTER, *PFS_OPERATIONS_FILTER;

#define OW_FS_OPERATIONS_FILTER_VERSION (1)

typedef struct _FS_OPERATIONS {
    
    //
    // General file system attributes
    //

    UNICODE_STRING    FsName;               // name of file system driver
    ULONG             Flags;                // flags
    ULONG             MaximumIoSize;        // maximum I/O transfer allowed, 0 = default
    ULONG             Attributes;           // FSD Attributes - OW_FS_ATTR_* listed below
    ULONG             ReadAheadSize;        // Maximum read-ahead size
    ULONG             WriteBehindSize;      // Maximum write-behind size
    ULONG             DirReadSize;          // Maximum read size for directory enumeration, 0 = default
    PVOID             UpcaseTable;          // FSD's upper case table, 0 = default
    ULONG             AlignmentRequirement; // e.g. sector size -1 (511 for a 512 byte sector.)
    ULONG             Reserved[3];          // Reserved for future use

    //
    // FSDK 2.0 Optional filter interface
    //
    PFS_OPERATIONS_FILTER      FilterTable;

    //
    // File System Driver Operations
    //
    
    FS_MOUNT                  Mount;
    FS_VERIFY                 Verify;
    FS_UNMOUNT                Unmount;
    FS_FSCTRL                 FsCtrl;
    FS_GET_QUOTA              GetQuota;
    FS_SET_QUOTA              SetQuota;
    FS_GET_VOLUME_ATTRIBUTES  GetVolAttributes;
    FS_SET_VOLUME_ATTRIBUTES  SetVolAttributes;

    //
    // File operations
    //

    FS_LOOKUP                  Lookup;
    FS_LOOKUP_BY_ID            LookupById;
    FS_LOOKUP_PATH             LookupPath;
    FS_SET_ACCESS_MODE         SetAccessMode;
    FS_RELEASE                 Release;
    FS_CREATE                  Create;
    FS_DELETE                  Delete;
    FS_RENAME                  Rename;
    FS_ACCESS                  Access;
    FS_GET_SECURITY            GetSecurity;
    FS_SET_SECURITY            SetSecurity;
    FS_LINK                    Link;
    FS_READ_SYMBOLIC_LINK      ReadSymbolicLink;
    FS_WRITE_SYMBOLIC_LINK     WriteSymbolicLink;
    FS_GET_ATTRIBUTES          GetAttributes;
    FS_SET_ATTRIBUTES          SetAttributes;
    FS_FLUSH                   Flush;
    FS_SET_LENGTH              SetLength;
    FS_GET_EA                  GetEA;
    FS_SET_EA                  SetEA;
    FS_READ                    Read;
    FS_WRITE                   Write;
    FS_IOCTL                   Ioctl;
    FS_READ_DIRECTORY          ReadDirectory;
    FS_READ_STREAM_INFORMATION ReadStreamInformation;

    //
    // Network operations
    //

    FS_QUERY_PATH             QueryPath;
    FS_CONNECT                Connect;
    FS_DISCONNECT             Disconnect;
    FS_UNC_ROOT               UncRoot;

    FS_LOCK                   Lock;
    FS_UNLOCK                 Unlock;

    FS_GET_UNC_VOLUME_ATTRIBUTES GetUncVolAttributes;
    FS_CHECK_LOCK                CheckLock;

    FS_UPDATE_SHARE_ACCESS    UpdateShareAccess;
    FS_REMOVE_SHARE_ACCESS    RemoveShareAccess;

    //
    // FSDK 2.0 New operations
    //

    FS_DELETE2                 Delete2;
    FS_READ_DIRECTORY2         ReadDirectory2;
    FS_RENAME2                 Rename2;

    FS_OPEN                    Open;
    FS_CLOSE                   Close;

    FS_READ_COMPRESSED         ReadCompressed;
    FS_WRITE_COMPRESSED        WriteCompressed;

    FS_READ                    PagefileRead;
    FS_WRITE                   PagefileWrite;

    FS_CLEAR                   Clear;

    FS_GET_NAMES               GetNames;

    //
    // Newer directory entry format
    //
    FS_READ_DIRECTORY3         ReadDirectory3;

    //
    // Open by Object ID
    //
    FS_LOOKUP_BY_OBJECT_ID     LookupByObjectId;

    //
    // Shutdown in progress
    //
    FS_SHUTDOWN                Shutdown;

    //
    // Space for future new functions
    //
    FS_RESERVED_FUNCTION       ReservedFunctions[9];

} FS_OPERATIONS, *PFS_OPERATIONS;

#define OSRFS_MINIMUM_IO_TRANSFER_SIZE 65536

//
// Access mode definitions.  These are now bits.
//

#define FS_ACCESS_NONE       (0x00)
#define FS_ACCESS_READ       (0x01)
#define FS_ACCESS_WRITE      (0x02)
#define FS_ACCESS_SEQUENTIAL (0x04)
#define FS_ACCESS_RANDOM     (0x08)
#define FS_ACCESS_TEMPORARY  (0x10)
#define FS_ACCESS_ALL        (FS_ACCESS_READ|FS_ACCESS_WRITE)

//
// FSD Attributes (for FS_OPERATIONS vector)
//

#define OW_FS_ATTR_NO_DIR_CACHE       0x00000001  // directory caching is disabled
#define OW_FS_ATTR_NO_FILE_CACHE      0x00000002  // file caching is disabled
#define OW_FS_ATTR_NO_OPLOCKS         0x00000004  // no oplock support
#define OW_FS_ATTR_COMPRESSION        0x00000008  // requests FSDK compression support
#define OW_FS_ATTR_SHORT_NAMES        0x00000010  // FSDK generates and FSD supports short file names
#define OW_FS_ATTR_WILDCARDS_LEGAL    0x00000020  // indicates wild cards are legal in file names
#define OW_FS_ATTR_NP_DELETE_DRIVE    0x00000040  // indicates the NP for a network FSD will delete the drive letter
#define OW_FS_ATTR_UNLOCKED_VOLUME_IO 0x00000080  // indicates the FSD allows unlocked DASD IO (not recommended.)
#define OW_FS_ATTR_SIMULATE_MOUNT     0x00000100  // simulate media mount for pseudo file system
#define OW_FS_ATTR_FSD_SHORT_NAMES    0x00000200  // FSD generates and supports short file names
#define OW_FS_ATTR_FSD_ACCESS_CHK     0x00000400  // FSD does all access checking (only valid on Pseudo FS)
#define OW_FS_ATTR_SAFE_REMOVAL       0x00000800  // FSD wants media user to remove device via safe removal otherwise
                                                  //   user can just yank device....
//
// Define this bit for Backward compatibility

//
// FS_LOOKUP now supplies a hint as to why the FSDK is doing a lookup.
// the hint is supplied as the value of the Type field. (In other words,
// if you dereference Type, it should be one of:
//
#define OW_LOOKUP_TYPE_NORMAL        0x10 // a user create for this filename
#define OW_LOOKUP_TYPE_RENAME_TARGET 0x11 // a rename operation to this filename
#define OW_LOOKUP_TYPE_TRAVERSE      0x12 // a directory traverse operation
#define OW_LOOKUP_TYPE_SHORTNAME     0x13 // a shortname generation operation
#define OW_LOOKUP_TYPE_LINK_TARGET   0x14 // a link operation to this filename

//
// note that these types are obliged to not overlap the set of valid return values
// for the type field so that the wrapper can detect a bad programmer on device
// fault. If OW_HANDLE types change, this have to change too.
//




#endif // __OSR_WRAPPER_H__
